package com.sf.service.gateway.service.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.http.HttpUtil;
import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONObject;
import com.sf.common.constant.HttpStatus;
import com.sf.common.exception.ServiceException;
import com.sf.common.utils.JSONUtils;
import com.sf.service.index.utils.AppUtils;
import com.sf.common.utils.DateUtils;
import com.sf.common.utils.SecurityUtils;
import com.sf.service.gateway.domain.*;
import com.sf.service.gateway.enums.EncryptionAlgorithm;
import com.sf.service.gateway.enums.GatewayDataStatus;
import com.sf.service.gateway.enums.GatewayServiceModel;
import com.sf.service.gateway.enums.GatewayStrategyType;
import com.sf.service.gateway.mapper.GatewayConfigMapper;
import com.sf.service.gateway.service.*;
import com.sf.vertx.api.pojo.*;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

import javax.annotation.Resource;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 功能描述:
 *
 * @author a_kun
 * @date 2024/4/25 17:24
 */
@Slf4j
@Service
public class GatewayConfigServiceImpl implements IGatewayConfigService {

    @Value("${sac.gateway.host:}")
    private String sacGatewayHost;

    @Value("${sac.gateway.port:5566}")
    private Integer sacGatewayPort;

    private static final String SYNC_CONFIG_PATH = "/vertx/app/config";

    private static final Long DEFAULT_API_TIMEOUT = 3000L;


    @Resource
    private GatewayConfigMapper gatewayConfigMapper;

    @Resource
    private IGatewayServerService gatewayServerService;

    @Resource
    private IGatewayRouteService gatewayRouteService;

    @Resource
    private IGatewayInterfaceInfoService gatewayInterfaceInfoService;

    @Resource
    private IGatewayStrategyService gatewayStrategyService;

    @Override
    public GatewayConfig selectGatewayConfigByAppCode(String appCode) {
        if (StringUtils.hasText(appCode)) {
            return gatewayConfigMapper.selectByAppCode(appCode);
        }
        return null;
    }

    @Override
    public void updateGatewayConfig(GatewayConfig gatewayConfig) {
        checkInsertOrUpdateGatewayConfig(gatewayConfig);
        String appCode = AppUtils.getAppCodeFromRequestHeader();
        gatewayConfig.setAppCode(appCode);
        String username = SecurityUtils.getUsername();
        Date nowDate = DateUtils.getNowDate();
        GatewayConfig inDbConfig = gatewayConfigMapper.selectByAppCode(appCode);
        if (inDbConfig == null) {
            gatewayConfig.setCreated(username);
            gatewayConfig.setModified(username);
            gatewayConfig.setCreateTime(nowDate);
            gatewayConfig.setUpdateTime(nowDate);
            gatewayConfigMapper.insert(gatewayConfig);
        } else {
            gatewayConfig.setModified(username);
            gatewayConfig.setCreateTime(nowDate);
            gatewayConfig.setUpdateTime(nowDate);
            gatewayConfigMapper.updateByAppCodeSelective(gatewayConfig);
        }
    }

    @Override
    public void synchronization(String appCode) {
        Assert.hasText(sacGatewayHost, "网关地址未配置");
        Assert.notNull(sacGatewayPort, "网关端口未配置");
        String urlStr = "http://" + sacGatewayHost + ":" + sacGatewayPort + SYNC_CONFIG_PATH;
        AppConfig appConfig = new AppConfig();
        appConfig.setAppCode(appCode);
        GatewayConfig gatewayConfig = selectGatewayConfigByAppCode(appCode);
        if (gatewayConfig != null) {
            appConfig.setDataSecurity(initDataSecurity(gatewayConfig));
            appConfig.setApiCurrentLimitingConfig(initApiCurrentLimitStrategy(gatewayConfig));
            appConfig.setAppCurrentLimitingConfig(initAppCurrentLimitStrategy(gatewayConfig));
            // appConfig.setAdvancedConfig();
        }
        appConfig.setService(initService(appCode));
        try {
            String body = JSON.toJSONString(appConfig);
            log.info("sync gateway body : {}", body);
            String response = HttpUtil.post(urlStr, body);
            log.info("sync gateway response : {}", response);
            if (StringUtils.hasText(response)) {
                JSONObject jsonObject = JSON.parseObject(response);
                if (jsonObject.getIntValue("code") == HttpStatus.SUCCESS) {
                    return;
                }
            }
        } catch (Exception e) {
            log.error("网关配置同步失败:", e);
            throw new ServiceException("网关配置同步失败，请稍后重试");
        }
        throw new ServiceException("网关配置同步失败，请稍后重试");
    }

    private List<SacService> initService(String appCode) {
        List<SacService> service = Collections.emptyList();
        GatewayServer queryServer = new GatewayServer();
        queryServer.setAppCode(appCode);
        queryServer.setStatus(GatewayDataStatus.ENABLE.getCode());
        List<GatewayServer> serverList = gatewayServerService.selectGatewayServerList(queryServer);
        if (CollUtil.isNotEmpty(serverList)) {
            // 查询接口
            List<GatewayInterfaceInfo> interfaceInfoList = gatewayInterfaceInfoService.selectGatewayInterfaceInfoByServerIds(serverList.stream().map(GatewayServer::getId).collect(Collectors.toSet()));
            // 启用的
            interfaceInfoList = interfaceInfoList.stream().filter(api -> GatewayDataStatus.ENABLE.getCode().equals(api.getStatus())).collect(Collectors.toList());
            if (CollUtil.isEmpty(interfaceInfoList)) {
                // 没有配置接口，直接结束
                return service;
            }
            Map<Long, List<GatewayInterfaceInfo>> groupByServerInterfaceMap = interfaceInfoList.stream().collect(Collectors.groupingBy(GatewayInterfaceInfo::getServerId));
            Set<Long> interfaceIds = interfaceInfoList.stream().map(GatewayInterfaceInfo::getId).collect(Collectors.toSet());
            List<GatewayInterfaceLinkStrategy> linkList = gatewayInterfaceInfoService.selectGatewayInterfaceLinkStrategyByInterfaceIds(interfaceIds);
            Map<Long, List<GatewayInterfaceLinkStrategy>> groupByInterfaceLinkMap = linkList.stream().collect(Collectors.groupingBy(GatewayInterfaceLinkStrategy::getInterfaceId));
            // 如果是路由模式的服务，查询对应的路由服务信息
            Collection<Long> routeIds = serverList.stream().filter(item -> GatewayServiceModel.ROUTE.getCode().equals(item.getServiceModel()))
                    .map(GatewayServer::getRouteId)
                    .collect(Collectors.toSet());
            List<GatewayRoute> routeList = gatewayRouteService.selectGatewayRouteByIds(routeIds);
            Map<Long, GatewayRoute> routeMap = routeList.stream().collect(Collectors.toMap(GatewayRoute::getId, Function.identity()));
            // 查询接口策略
            List<GatewayStrategy> strategyList = gatewayStrategyService.selectGatewayStrategyByIds(linkList.stream().map(GatewayInterfaceLinkStrategy::getStrategyId).collect(Collectors.toSet()));
            // 启用的
            strategyList = strategyList.stream().filter(strategy -> GatewayDataStatus.ENABLE.getCode().equals(strategy.getStatus())).collect(Collectors.toList());
            Map<Long, GatewayStrategy> strategyMap = strategyList.stream().collect(Collectors.toMap(GatewayStrategy::getId, Function.identity()));
            // mock
            List<GatewayInterfaceMock> mockList = gatewayInterfaceInfoService.selectGatewayInterfaceMockListByInterfaceIds(interfaceIds);
            Map<Long, List<GatewayInterfaceMock>> groupByInterfaceMockMap = mockList.stream()
                    .filter(mock -> GatewayDataStatus.ENABLE.getCode().equals(mock.getStatus()))
                    .collect(Collectors.groupingBy(GatewayInterfaceMock::getInterfaceId));
            return serverList.stream()
                    .filter(server -> CollUtil.isNotEmpty(groupByServerInterfaceMap.getOrDefault(server.getId(), Collections.emptyList())))
                    .map(server -> {
                        long serviceTimeout = server.getTimeout() != null ? server.getTimeout().longValue() : DEFAULT_API_TIMEOUT;
                        SacService sacService = new SacService();
                        sacService.setServiceName(server.getServerName());
                        sacService.setServiceModel(server.getServiceModel());
                        sacService.setServiceType(server.getServiceType());
                        sacService.setServerAddress(convertToServerAddress(server.getServerAddress()));
                        List<GatewayInterfaceInfo> currentServerApiList = groupByServerInterfaceMap.getOrDefault(server.getId(), Collections.emptyList());
                        List<ApiConfig> apiConfigList = currentServerApiList.stream().map(api -> {
                            List<GatewayInterfaceLinkStrategy> currentApiLinkStrategy = groupByInterfaceLinkMap.getOrDefault(api.getId(), Collections.emptyList());
                            List<Strategy> currentApiStrategyList = currentApiLinkStrategy.stream()
                                    .map(link -> convertToStrategy(strategyMap.get(link.getStrategyId())))
                                    .collect(Collectors.toList());
                            ApiConfig apiConfig = new ApiConfig();
                            apiConfig.setApiCode(api.getApiCode());
                            apiConfig.setUri(api.getInterfacePath());
                            apiConfig.setMethod(api.getRequestMethod());
                            apiConfig.setTimeout(api.getTimeout() != null ? api.getTimeout().longValue() : serviceTimeout);
                            apiConfig.setStrategy(currentApiStrategyList);
                            // Mock
                            if (GatewayDataStatus.ENABLE.getCode().equals(api.getMockStatus())) {
                                apiConfig.setMockDefaultHttpStatus(api.getMockDefaultHttpStatus());
                                apiConfig.setMockDefaultResponse(api.getMockDefaultResponse());
                                List<GatewayInterfaceMock> currentMocks = groupByInterfaceMockMap.getOrDefault(api.getId(), Collections.emptyList());
                                List<MockExpectation> mockExpectations = currentMocks.stream()
                                        .sorted(Comparator.comparingInt(GatewayInterfaceMock::getOrderNum))
                                        .map(mock ->{
                                            MockExpectation mockExpectation = new MockExpectation();
                                            mockExpectation.setHttpStatus(mock.getHttpStatus());
                                            mockExpectation.setMockResponse(mock.getMockResponse());
                                            mockExpectation.setMatchConditions(JSON.parseArray(mock.getMatchConditions(),MockMatchCondition.class));
                                            return mockExpectation;
                                        })
                                        .collect(Collectors.toList());
                                apiConfig.setMockExpectations(mockExpectations);
                            }
                            return apiConfig;
                        }).collect(Collectors.toList());
                        sacService.setApiConfig(apiConfigList);
                        // 设置路由
                        if (GatewayServiceModel.ROUTE.getCode().equals(server.getServiceModel())) {
                            GatewayRoute route = routeMap.get(server.getRouteId());
                            Router routeConfig = new Router();
                            routeConfig.setRouteType(route.getRouteType());
                            List<GatewayRoute.RouteContent> routeContents = JSON.parseArray(route.getRouteContent(), GatewayRoute.RouteContent.class);
                            routeConfig.setRouteContent(routeContents.stream().map(content -> {
                                RouteContent routeContent = new RouteContent();
                                routeContent.setServerAddress(convertToServerAddress(content.getServerAddress()));
                                routeContent.setWeight(content.getWeight());
                                routeContent.setHeaderKey(content.getHeaderKey());
                                routeContent.setHeaderValues(content.getHeaderValues());
                                routeContent.setMatchType(content.getMatchType());
                                return routeContent;
                            }).collect(Collectors.toList()));
                            sacService.setRouteConfig(routeConfig);
                        }
                        return sacService;
                    }).collect(Collectors.toList());
        }
        return service;
    }

    private Strategy convertToStrategy(GatewayStrategy currentStrategy) {
        Strategy strategy = new Strategy();
        strategy.setType(currentStrategy.getStrategyType());
        switch (GatewayStrategyType.getByCode(currentStrategy.getStrategyType())) {
            case CURRENT_LIMITING:
                strategy.setThreshold(currentStrategy.getCurrentLimitingThreshold());
                strategy.setTimeWindow(currentStrategy.getCurrentLimitingTimeWindow());
                strategy.setDefaultResponse(currentStrategy.getCurrentLimitingResponse());
                break;
            case CIRCUIT_BREAKER:
                strategy.setThreshold(currentStrategy.getCircuitBreakerThreshold());
                strategy.setTimeWindow(currentStrategy.getCircuitBreakerTimeWindow());
                strategy.setRecovery_interval(currentStrategy.getCircuitBreakerRecoveryInterval());
                strategy.setDefaultResponse(currentStrategy.getCircuitBreakerResponse());
                break;
            case AUTHENTICATION:
            case PERMISSION:
            default:
                break;

        }
        return strategy;
    }

    private ServerAddress convertToServerAddress(String serverAddress) {
        if (StringUtils.hasText(serverAddress)) {
            try {
                URL url = new URL(serverAddress);
                ServerAddress address = new ServerAddress();
                address.setAddress(serverAddress);
                address.setProtocol(url.getProtocol());
                address.setHost(url.getHost());
                address.setPort(url.getPort() == -1 ? 80 : url.getPort());
                address.setPath(url.getPath());
                return address;
            } catch (MalformedURLException ignored) {
            }
        }
        return null;
    }

    private Strategy initAppCurrentLimitStrategy(GatewayConfig gatewayConfig) {
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayConfig.getAppCurrentLimitingStatus())) {
            Strategy appCurrentLimitStrategy = new Strategy();
            appCurrentLimitStrategy.setType(GatewayStrategyType.CURRENT_LIMITING.getCode());
            appCurrentLimitStrategy.setThreshold(gatewayConfig.getAppCurrentLimitingThreshold());
            appCurrentLimitStrategy.setTimeWindow(gatewayConfig.getAppCurrentLimitingTimeWindow());
            appCurrentLimitStrategy.setDefaultResponse(gatewayConfig.getAppCurrentLimitingDefaultResponse());
            return appCurrentLimitStrategy;
        }
        return null;
    }

    private Strategy initApiCurrentLimitStrategy(GatewayConfig gatewayConfig) {
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayConfig.getApiCurrentLimitingStatus())) {
            Strategy apiCurrentLimitStrategy = new Strategy();
            apiCurrentLimitStrategy.setType(GatewayStrategyType.CURRENT_LIMITING.getCode());
            apiCurrentLimitStrategy.setThreshold(gatewayConfig.getApiCurrentLimitingThreshold());
            apiCurrentLimitStrategy.setTimeWindow(gatewayConfig.getApiCurrentLimitingTimeWindow());
            apiCurrentLimitStrategy.setDefaultResponse(gatewayConfig.getApiCurrentLimitingDefaultResponse());
            return apiCurrentLimitStrategy;
        }
        return null;
    }

    /**
     * 初始化数据加密配置
     *
     * @param gatewayConfig
     */
    private DataSecurity initDataSecurity(GatewayConfig gatewayConfig) {
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayConfig.getDataEncryptionStatus())) {
            DataSecurity dataSecurity = new DataSecurity();
            dataSecurity.setAlgorithm(gatewayConfig.getEncryptionAlgorithm());
            dataSecurity.setPublicKey(gatewayConfig.getPublicKey());
            dataSecurity.setPrivateKey(gatewayConfig.getPrivateKey());
            return dataSecurity;
        }
        return null;
    }

    private void checkInsertOrUpdateGatewayConfig(GatewayConfig gatewayConfig) {
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayConfig.getDataEncryptionStatus())) {
            EncryptionAlgorithm encryptionAlgorithm = EncryptionAlgorithm.getByCode(gatewayConfig.getEncryptionAlgorithm());
            Assert.notNull(encryptionAlgorithm, "不支持的加密算法");
            Assert.hasText(gatewayConfig.getPrivateKey(), "密钥内容不能为空");
            if (EncryptionAlgorithm.RSA.equals(encryptionAlgorithm)) {
                Assert.hasText(gatewayConfig.getPublicKey(), "公钥内容不能为空");
            }
        }
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayConfig.getApiCurrentLimitingStatus())) {
            Assert.notNull(gatewayConfig.getApiCurrentLimitingThreshold(), "限流值不能为空");
            Assert.isTrue(gatewayConfig.getApiCurrentLimitingThreshold() >= 0, "限流值>=0");
            // 页面没有,这里给个默认值
            if (gatewayConfig.getApiCurrentLimitingTimeWindow() == null) {
                gatewayConfig.setApiCurrentLimitingTimeWindow(1);
            } else {
                Assert.isTrue(gatewayConfig.getApiCurrentLimitingTimeWindow() >= 0, "限流窗口值>=0");
            }
            Assert.isTrue(JSONUtils.isJSONObject(gatewayConfig.getApiCurrentLimitingDefaultResponse()), "请输入正确的JSON格式限流响应信息");
        }
        if (GatewayDataStatus.ENABLE.getCode().equals(gatewayConfig.getAppCurrentLimitingStatus())) {
            Assert.notNull(gatewayConfig.getAppCurrentLimitingThreshold(), "限流值不能为空");
            Assert.isTrue(gatewayConfig.getAppCurrentLimitingThreshold() >= 0, "限流值>=0");
            // 页面没有,这里给个默认值
            if (gatewayConfig.getAppCurrentLimitingTimeWindow() == null) {
                gatewayConfig.setAppCurrentLimitingTimeWindow(1);
            } else {
                Assert.isTrue(gatewayConfig.getAppCurrentLimitingTimeWindow() >= 0, "限流窗口值>=0");
            }
            Assert.isTrue(JSONUtils.isJSONObject(gatewayConfig.getAppCurrentLimitingDefaultResponse()), "请输入正确的JSON格式限流响应信息");
        }
    }
}
