/*
 * Copyright (c) 2011-2020 Contributors to the Eclipse Foundation
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
 * which is available at https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
 */
package io.vertx.httpproxy.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.function.BiFunction;

import cn.hutool.core.util.StrUtil;
import com.sf.vertx.enums.RequestMethod;
import org.apache.commons.lang3.StringUtils;

import com.sf.vertx.constans.SacErrorCode;
import com.sf.vertx.handle.AppConfigHandler;
import com.sf.vertx.utils.ProxyTool;

import io.vertx.circuitbreaker.CircuitBreaker;
import io.vertx.circuitbreaker.HalfOpenCircuitException;
import io.vertx.circuitbreaker.OpenCircuitException;
import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpClientRequest;
import io.vertx.core.http.HttpClientResponse;
import io.vertx.core.http.HttpHeaders;
import io.vertx.core.http.HttpMethod;
import io.vertx.core.http.HttpServerRequest;
import io.vertx.core.http.HttpServerResponse;
import io.vertx.core.impl.NoStackTraceThrowable;
import io.vertx.core.json.JsonObject;
import io.vertx.core.net.NetSocket;
import io.vertx.core.net.SocketAddress;
import io.vertx.ext.web.RoutingContext;
import io.vertx.ext.web.client.HttpRequest;
import io.vertx.ext.web.client.WebClient;
import io.vertx.ext.web.handler.HttpException;
import io.vertx.httpproxy.Body;
import io.vertx.httpproxy.HttpProxy;
import io.vertx.httpproxy.ProxyContext;
import io.vertx.httpproxy.ProxyInterceptor;
import io.vertx.httpproxy.ProxyOptions;
import io.vertx.httpproxy.ProxyRequest;
import io.vertx.httpproxy.ProxyResponse;
import io.vertx.httpproxy.cache.CacheOptions;
import io.vertx.httpproxy.spi.cache.Cache;
import lombok.extern.slf4j.Slf4j;

import static org.apache.commons.lang3.StringUtils.EMPTY;

@Slf4j
public class ReverseProxy implements HttpProxy {

	private final HttpClient client;
	private final boolean supportWebSocket;
	private BiFunction<HttpServerRequest, HttpClient, Future<HttpClientRequest>> selector = (req, client) -> Future
			.failedFuture("No origin available");
	private final List<ProxyInterceptor> interceptors = new ArrayList<>();
	private RoutingContext ctx;
	private WebClient mainWebClient;

	public ReverseProxy(ProxyOptions options, HttpClient client) {
		CacheOptions cacheOptions = options.getCacheOptions();
		if (cacheOptions != null) {
			Cache<String, Resource> cache = cacheOptions.newCache();
			addInterceptor(new CachingFilter(cache));
		}
		this.client = client;
		this.supportWebSocket = options.getSupportWebSocket();
	}

	@Override
	public HttpProxy originRequestProvider(
			BiFunction<HttpServerRequest, HttpClient, Future<HttpClientRequest>> provider) {
		selector = provider;
		return this;
	}

	@Override
	public HttpProxy addInterceptor(ProxyInterceptor interceptor) {
		interceptors.add(interceptor);
		return this;
	}

	@Override
	public void handle(WebClient mainWebClient, RoutingContext ctx) {
		// TODO 改造了这个地方
		this.ctx = ctx;
		this.mainWebClient = mainWebClient;
		handle(ctx.request());
	}

	@Override
	public void handle(HttpServerRequest request) {
		ProxyRequest proxyRequest = ProxyRequest.reverseProxy(request);

		// Encoding sanity check
		Boolean chunked = HttpUtils.isChunked(request.headers());
		if (chunked == null) {
			end(proxyRequest, 400);
			return;
		}

		// WebSocket upgrade tunneling
		if (supportWebSocket && io.vertx.core.http.impl.HttpUtils.canUpgradeToWebSocket(request)) {
			handleWebSocketUpgrade(proxyRequest);
			return;
		}

		Proxy proxy = new Proxy(proxyRequest);
		proxy.filters = interceptors.listIterator();
		proxy.sendRequest().compose(proxy::sendProxyResponse);
	}

	private void handleWebSocketUpgrade(ProxyRequest proxyRequest) {
		HttpServerRequest proxiedRequest = proxyRequest.proxiedRequest();
		resolveOrigin(proxiedRequest).onComplete(ar -> {
			if (ar.succeeded()) {
				HttpClientRequest request = ar.result();
				request.setMethod(HttpMethod.GET);
				request.setURI(proxiedRequest.uri());
				request.headers().addAll(proxiedRequest.headers());
				Future<HttpClientResponse> fut2 = request.connect();
				proxiedRequest.handler(request::write);
				proxiedRequest.endHandler(v -> request.end());
				proxiedRequest.resume();
				fut2.onComplete(ar2 -> {
					if (ar2.succeeded()) {
						HttpClientResponse proxiedResponse = ar2.result();
						if (proxiedResponse.statusCode() == 101) {
							HttpServerResponse response = proxiedRequest.response();
							response.setStatusCode(101);
							response.headers().addAll(proxiedResponse.headers());
							Future<NetSocket> otherso = proxiedRequest.toNetSocket();
							otherso.onComplete(ar3 -> {
								if (ar3.succeeded()) {
									NetSocket responseSocket = ar3.result();
									NetSocket proxyResponseSocket = proxiedResponse.netSocket();
									responseSocket.handler(proxyResponseSocket::write);
									proxyResponseSocket.handler(responseSocket::write);
									responseSocket.closeHandler(v -> proxyResponseSocket.close());
									proxyResponseSocket.closeHandler(v -> responseSocket.close());
								} else {
									// Find reproducer
									System.err.println("Handle this case");
									ar3.cause().printStackTrace();
								}
							});
						} else {
							// Rejection
							proxiedRequest.resume();
							end(proxyRequest, proxiedResponse.statusCode());
						}
					} else {
						proxiedRequest.resume();
						end(proxyRequest, 502);
					}
				});
			} else {
				proxiedRequest.resume();
				end(proxyRequest, 502);
			}
		});
	}

	/***
	 * 返回错误
	 * 
	 * @param proxyRequest
	 * @param sc
	 */
	private void end(ProxyRequest proxyRequest, int sc) {
		JsonObject json = SacErrorCode.returnErrorMsg(sc);
		proxyRequest.response().release().setStatusCode(500).putHeader("content-type", "application/json")
				.putHeader(AppConfigHandler.sacResponseHeaderKey(), String.valueOf(sc))
				.putHeader(HttpHeaders.CONTENT_LENGTH, String.valueOf(json.size())).setBody(Body.body(json.toBuffer()))
				.send();
	}

	private Future<HttpClientRequest> resolveOrigin(HttpServerRequest proxiedRequest) {
		return selector.apply(proxiedRequest, client);
	}

	private class Proxy implements ProxyContext {

		private final ProxyRequest request;
		private ProxyResponse response;
		private final Map<String, Object> attachments = new HashMap<>();
		private ListIterator<ProxyInterceptor> filters;

		private Proxy(ProxyRequest request) {
			this.request = request;
		}

		@Override
		public void set(String name, Object value) {
			attachments.put(name, value);
		}

		@Override
		public <T> T get(String name, Class<T> type) {
			Object o = attachments.get(name);
			return type.isInstance(o) ? type.cast(o) : null;
		}

		@Override
		public ProxyRequest request() {
			return request;
		}

		@Override
		public Future<ProxyResponse> sendRequest() {
			if (filters.hasNext()) {
				ProxyInterceptor next = filters.next();
				return next.handleProxyRequest(this);
			} else {
				return sendProxyRequest(request);
			}
		}

		@Override
		public ProxyResponse response() {
			return response;
		}

		@Override
		public Future<Void> sendResponse() {
			if (filters.hasPrevious()) {
				ProxyInterceptor filter = filters.previous();
				return filter.handleProxyResponse(this);
			} else {
				return response.send();
			}
		}

		private Future<ProxyResponse> breakerCall(CircuitBreaker circuitBreaker
				, ProxyRequest proxyRequest
				, HttpRequest<Buffer> requestBuffer
				, String body
		) {
			String appCode = proxyRequest.headers().get(AppConfigHandler.getAppCodeHeaderKey());
			return Future.future(p -> {
				circuitBreaker.executeWithFallback(promise -> {
					requestBuffer.sendJson(body, h -> {
								if (h.succeeded()) {
									log.info("==========uri:{},response http code:{}", proxyRequest.getURI(),
											h.result().statusCode());
									if (h.result().statusCode() == 200) {
										// promise.complete();
										promise.complete("1");
									} else {
										// Throwable throwable = new Throwable("error port");
										// promise.fail(throwable);
										promise.fail("2");
									}
									// 释放资源
									proxyRequest.release();
									JsonObject responseData = h.result().bodyAsJsonObject();
									log.info("responseData:{}", responseData);
									Buffer buffer;
									// 加密
									if (AppConfigHandler.isDataSecurity(appCode)) {
										String dataStr = AppConfigHandler.bodyEncrypt(responseData.toString(), appCode);
										log.info("encrypt dataStr:{}", dataStr);
										buffer = Buffer.buffer(dataStr);
									}else {
										buffer = responseData.toBuffer();
									}
									ProxyResponse proxyResponse = proxyRequest.response()
											.setStatusCode(200)
											.putHeader("content-type", "application/json")
											.setBody(Body.body(buffer));
									p.complete(proxyResponse);
								} else {
									// end(proxyRequest, 502);
									// Throwable throwable = new Throwable("error port");
									// promise.fail(throwable);
									promise.fail("3");
								}
							});
				}, v -> {
					// 需要传递当前状态half-open , close, 还是统计失败次数
					log.info(circuitBreaker.name() + " executed when the circuit is opened:{}", v.getMessage());
					if (v instanceof HalfOpenCircuitException) {
						log.info(circuitBreaker.name() + " half open circuit");
						return v.getMessage();
					} else if (v instanceof OpenCircuitException) {
						log.info(circuitBreaker.name() + " open circuit");
					} else if (v instanceof NoStackTraceThrowable) {
						log.info(circuitBreaker.name() + " close circuit");
						return v.getMessage();
					}
					return "3";
				}, ar -> {
					log.info(circuitBreaker.name() + " interface failed result.{} ", ar);
					if (StringUtils.equals(ar.result(), "3")) { // 全开,熔断
						end(proxyRequest, 10016);
					}
				});
			});
		}


		private Future<ProxyResponse> normalCall(ProxyRequest proxyRequest, HttpRequest<Buffer> requestBuffer, String body) {
			String appCode = proxyRequest.headers().get(AppConfigHandler.getAppCodeHeaderKey());
			return Future.future(p -> {
				requestBuffer.sendJson(body,
						h -> {
							if (h.succeeded()) {
								log.info("==========uri:{},response http code:{}", proxyRequest.getURI(),
										h.result().statusCode());
								if (h.result().statusCode() == 200) {
									// 释放资源
									proxyRequest.release();
									JsonObject responseData = h.result().bodyAsJsonObject();
									log.info("responseData:{}", responseData);
									Buffer bodyBuffer = responseData.toBuffer();
									// 加密
									if (AppConfigHandler.isDataSecurity(appCode)) {
										String dataStr = AppConfigHandler.bodyEncrypt(responseData.toString(), appCode);
										log.info("encrypt dataStr:{}", dataStr);
										bodyBuffer = Buffer.buffer(dataStr);
									}
									ProxyResponse proxyResponse = proxyRequest.response().setStatusCode(200)
											.putHeader("content-type", "application/json").setBody(Body.body(bodyBuffer));
									p.complete(proxyResponse);
								}
							} else {
								log.info("interface retrun error.{}", proxyRequest.getURI());
								end(proxyRequest, 502);
							}
						});
			});
		}


		private HttpRequest<Buffer> buildCallRequestBuffer(ProxyRequest proxyRequest,String body) {
			String appCode = proxyRequest.headers().get(AppConfigHandler.getAppCodeHeaderKey());
			String apiCode = proxyRequest.headers().get(AppConfigHandler.getApiCodeHeaderKey());
			String key = appCode + ":" + apiCode;
			SocketAddress socketAddress = ProxyTool.resolveOriginAddress(proxyRequest.proxiedRequest());
			HttpRequest<Buffer> requestBuffer;
			String requestURI = proxyRequest.getURI();
			switch (proxyRequest.getMethod().name()) {
			case "PUT":
				requestBuffer = mainWebClient.put(socketAddress.port(), socketAddress.host(), requestURI);
				break;
			case "DELETE":
				requestBuffer = mainWebClient.delete(socketAddress.port(), socketAddress.host(), requestURI);
				break;
			 case "HEAD":
				requestBuffer = mainWebClient.head(socketAddress.port(), socketAddress.host(), requestURI);
				break;
			case "GET":
				requestBuffer = mainWebClient.get(socketAddress.port(), socketAddress.host(), requestURI);
				break;
			default:
				requestBuffer = mainWebClient.post(socketAddress.port(), socketAddress.host(), requestURI);
				break;
			}
			requestBuffer.timeout(AppConfigHandler.getApicodeConfigTimeOut(key));
			requestBuffer.putHeaders(proxyRequest.headers());
			// 处理sac请求方式
			RequestMethod requestMethod = RequestMethod.getByCode(proxyRequest.getMethod().name());
			if (RequestMethod.GET.equals(requestMethod)
					|| RequestMethod.DELETE.equals(requestMethod)
					|| RequestMethod.HEAD.equals(requestMethod)){
				JsonObject jsonBody = new JsonObject(body);
				Map<String, String> queryParams = new HashMap<>();
				for (Map.Entry<String, Object> entry : jsonBody) {
					String val = entry.getValue() != null ? entry.getValue().toString() : EMPTY;
					queryParams.put(entry.getKey(),val);
					requestURI = StrUtil.replace(requestURI,"{" + entry.getKey() +"}",val);
				}
				requestBuffer.uri(requestURI);
				requestBuffer.queryParams().addAll(queryParams);
			}
			return requestBuffer;
		}

		private Future<ProxyResponse> sendProxyRequest(ProxyRequest proxyRequest) {
			// 判断<br/>
			// 1、是否配置全局加解密.<br/>
			// 2、apiCode 配置熔断
			String appCode = proxyRequest.headers().get(AppConfigHandler.getAppCodeHeaderKey());
			String apiCode = proxyRequest.headers().get(AppConfigHandler.getApiCodeHeaderKey());
			String contentType = proxyRequest.headers().get(HttpHeaders.CONTENT_TYPE);
			String keyCircuitBreaker = appCode + ":" + apiCode + ":" + "CIRCUIT_BREAKER";
			CircuitBreaker circuitBreaker = AppConfigHandler.getApiCodeCircuitBreaker(keyCircuitBreaker);
			if (AppConfigHandler.isAnalysisBody(appCode, apiCode, contentType)) {
				String body = AppConfigHandler.isDataSecurity(appCode) ? AppConfigHandler.bodyDecrypt(ctx.body().asString(), appCode):ctx.body().asString();
				HttpRequest<Buffer> requestBuffer = buildCallRequestBuffer(proxyRequest,body);
				try {
					 if (circuitBreaker != null) {
						return breakerCall(circuitBreaker, proxyRequest,requestBuffer,body);
					} else {
						return normalCall(proxyRequest,requestBuffer,body);
					}
				} catch (Exception e) {
					e.printStackTrace();
					int errorCode = 10014;
					if (e instanceof HttpException) {
						errorCode = ((HttpException) e).getStatusCode();
					}
					throw new HttpException(errorCode);
				}
			} else {
				Future<HttpClientRequest> f = resolveOrigin(proxyRequest.proxiedRequest());
				f.onFailure(err -> {
					log.info("error:", err);
					// Should this be done here ? I don't think so
					HttpServerRequest proxiedRequest = proxyRequest.proxiedRequest();
					proxiedRequest.resume();
					Promise<Void> promise = Promise.promise();
					proxiedRequest.exceptionHandler(promise::tryFail);
					proxiedRequest.endHandler(promise::tryComplete);
					promise.future().onComplete(ar2 -> {
						end(proxyRequest, 502);
					});
				});
				return f.compose(a -> sendProxyRequest(proxyRequest, a));
			}

		}

		private Future<ProxyResponse> sendProxyRequest(ProxyRequest proxyRequest, HttpClientRequest request) {
			String appCode = proxyRequest.headers().get(AppConfigHandler.getAppCodeHeaderKey());
			String apiCode = proxyRequest.headers().get(AppConfigHandler.getApiCodeHeaderKey());
			String key = appCode + ":" + apiCode;
			request.setTimeout(AppConfigHandler.getApicodeConfigTimeOut(key));
			Future<ProxyResponse> fut = proxyRequest.send(request);
			fut.onFailure(err -> {
				JsonObject json = SacErrorCode.returnErrorMsg(502);
				proxyRequest.response().release().setStatusCode(500).putHeader("content-type", "application/json")
				.putHeader(AppConfigHandler.sacResponseHeaderKey(), String.valueOf(502))
				.putHeader(HttpHeaders.CONTENT_LENGTH, String.valueOf(json.size())).setBody(Body.body(json.toBuffer()))
				.send();
				//proxyRequest.proxiedRequest().response().setStatusCode(502).end();
			});
			return fut;
		}

		private Future<Void> sendProxyResponse(ProxyResponse response) {

			this.response = response;

			// Check validity
			Boolean chunked = HttpUtils.isChunked(response.headers());
			if (chunked == null) {
				// response.request().release(); // Is it needed ???
				end(response.request(), 501);
				return Future.succeededFuture(); // should use END future here ???
			}

			return sendResponse();
		}


	}

}
